@@ -119,9 +119,25 @@ void cancel_drop(gboolean state)
     drop_cancelled = state;
 }
 
-static void clear_selection_list(void)
+void reselect_dnd_list(GtkTreeView *treeview){
+    GList *tmp;
+    GtkTreeSelection *selection = gtk_tree_view_get_selection(treeview);
+    for (tmp=selection_list;tmp;tmp=tmp->next){
+	selection_list_t *sl=tmp->data;
+        if(sl && gtk_tree_row_reference_valid(sl->reference)){
+           GtkTreePath *treepath = gtk_tree_row_reference_get_path(sl->reference);
+	   /*printf("DBG: reselecting %s\n",sl->en->path);*/
+	   gtk_tree_selection_select_path (selection,treepath);
+	   gtk_tree_path_free(treepath);
+	}
+    }
+}
+
+void clear_dnd_selection_list(void)
 {
     GList *tmp = selection_list;
+    if (dragging) return;
+    /*printf("DBG: clearing selection list!\n");*/
     while(tmp)
     {
 	selection_list_t *sl;
@@ -141,6 +157,35 @@ static void clear_selection_list(void)
     selection_list = NULL;
 }
 
+void clear_path_from_selection_list(GtkTreeView *treeview,GtkTreePath *treepath)
+{
+    GtkTreePath *path;
+    GtkTreeSelection *selection = gtk_tree_view_get_selection(treeview);
+    GList *tmp = selection_list;
+    if (!treepath || ! treeview) return;
+    gtk_tree_selection_unselect_path (selection,treepath);
+    for (tmp=selection_list;tmp;tmp=tmp->next)
+    {
+	selection_list_t *sl;
+	sl = (selection_list_t *) tmp->data;
+	path=gtk_tree_row_reference_get_path(sl->reference);
+	if (gtk_tree_path_compare(path,treepath)==0){
+	    selection_list = g_list_remove(selection_list,sl);
+	    g_free(sl);
+	    gtk_tree_path_free(path);
+	    return;
+	}
+        gtk_tree_path_free(path);
+    }
+    return;
+}
+
+void   on_drag_begin(GtkWidget *widget,GdkDragContext *drag_context,gpointer user_data){
+    	dragging = TRUE;	
+	/* this works better with plain motion-event */
+	/*reselect_dnd_list((GtkTreeView *)widget);   */
+}
+
 static gboolean valid_drop_site(GtkTreeView * treeview, gint x, gint y, tree_entry_t ** en, GtkTreeRowReference ** reference)
 {
     tree_details_t *tree_details = get_tree_details(treeview);
@@ -653,27 +698,42 @@ void on_drag_data(GtkWidget * widget, GdkDragContext * context, gint x, gint y,
     if (!IS_BOOKMARK_TYPE(en->type)) local_monitor(treeview,TRUE);
 }
 
+gboolean is_in_dnd_selection(GtkTreePath * path){
+	GList *tmp;
+	if (!path) return TRUE;
+	for (tmp=selection_list;tmp;tmp=tmp->next){
+		selection_list_t *sl=tmp->data;
+           	GtkTreePath *treepath = gtk_tree_row_reference_get_path(sl->reference);
+		if(gtk_tree_path_compare (path,treepath)==0){
+	           gtk_tree_path_free(treepath);
+		   return TRUE;
+		}
+	        gtk_tree_path_free(treepath);
+	}
+	return FALSE;
+}
+
 void get_dnd_selection(GtkTreeModel * treemodel, GtkTreePath * path, GtkTreeIter * iter, gpointer data)
 {
     GtkTreeView *treeview = (GtkTreeView *) data;
     GtkTreeSelection *selection = gtk_tree_view_get_selection(treeview);
     tree_entry_t *en;
 
-    if (drag_type==DRAG_TYPE_INCONSISTENT) return;
+   /* if (drag_type==DRAG_TYPE_INCONSISTENT) return;*/
 
     gtk_tree_model_get(treemodel, iter, ENTRY_COLUMN, &en, -1);
-    if (!IS_LOCAL_TYPE(en->type) && IS_ROOT_TYPE(en->type)) return;
     if(!IS_PATH(en->type) && !IS_NETTHING(en->subtype) )
     {
 	gtk_tree_selection_unselect_iter(selection, iter);
 	return;
     }
-    if((IS_PATH(en->type) && drag_type==DRAG_TYPE_NET) ||
+    if (!IS_LOCAL_TYPE(en->type) && IS_ROOT_TYPE(en->type)) return;
+    /*if((IS_PATH(en->type) && drag_type==DRAG_TYPE_NET) ||
        (IS_NETTHING(en->subtype) && drag_type==DRAG_TYPE_LOCAL)){
         drag_type=DRAG_TYPE_INCONSISTENT;
 	print_diagnostics(treeview,"xf_ERROR_ICON",strerror(E2BIG),"\n",NULL); 
 	return;
-    }
+    }*/
     if(IS_PATH(en->type)) drag_type=DRAG_TYPE_LOCAL;
     if(IS_NETTHING(en->subtype)) drag_type=DRAG_TYPE_NET;
 
@@ -685,10 +745,9 @@ void get_dnd_selection(GtkTreeModel * treemodel, GtkTreePath * path, GtkTreeIter
 	sl->en = en;
 	sl->reference = gtk_tree_row_reference_new(treemodel, path);
 	selection_list = g_list_append(selection_list, sl);
+/*	selection_list = g_list_prepend(selection_list, sl);*/
+	/*printf("DBG: added to selection list=%s\n",en->path);*/
     }
-    selection_len += (strlen(en->path) + 7);
-
-
     return;
 }
 
@@ -705,21 +764,17 @@ void on_drag_data_get(GtkWidget * widget, GdkDragContext * context, GtkSelection
     GtkTreeView *treeview = (GtkTreeView *) widget;
     GtkTreeSelection *selection = gtk_tree_view_get_selection(treeview);
     
-    if(dragging) return;
-    dragging = TRUE;
-
     if(!widget)	return;
-    if(selection_list)	clear_selection_list();
-
+    /*if(selection_list)	clear_dnd_selection_list();*/
     
-    drag_type=DRAG_TYPE_UNDEFINED;
-    gtk_tree_selection_selected_foreach(selection, get_dnd_selection, (gpointer) treeview);
-    if(!selection_list || !(drag_type&(DRAG_TYPE_LOCAL|DRAG_TYPE_NET))) 
+    /*gtk_tree_selection_selected_foreach(selection, get_dnd_selection, (gpointer) treeview);*/
+    /*drag_type=DRAG_TYPE_UNDEFINED;*/
+    if(!selection_list) return;
+   
+    if(!(drag_type&(DRAG_TYPE_LOCAL|DRAG_TYPE_NET))) 
 	    return;
 
 
-
-
     /* prepare data for the receiver */
     switch (info)
     {
@@ -734,12 +789,20 @@ void on_drag_data_get(GtkWidget * widget, GdkDragContext * context, GtkSelection
 	  default:
 		  g_assert_not_reached();
 	  case DRAG_TYPE_LOCAL:	    
+	    selection_len = 0;
+	    for (tmp = selection_list;tmp;tmp = tmp->next) {
+	        selection_list_t *sl;
+	        sl = (selection_list_t *) tmp->data;
+		if (!gtk_tree_row_reference_valid(sl->reference)) return;
+	        selection_len += (strlen(sl->en->path) + 7);
+	    }
 	    dnd_data = files = g_malloc(selection_len + 1);
 	    dnd_data[0] = '\0';
 	    for (tmp = selection_list;tmp;tmp = tmp->next)
 	    {
 		selection_list_t *sl;
 		sl = (selection_list_t *) tmp->data;
+		if (!gtk_tree_row_reference_valid(sl->reference)) return;
 		if(strncmp(sl->en->path, "tar:", strlen("tar:")) == 0)
 		{
 		    sprintf(files, "%s\r\n", sl->en->path);
@@ -758,6 +821,7 @@ void on_drag_data_get(GtkWidget * widget, GdkDragContext * context, GtkSelection
 	        int addlen;
 		selection_list_t *sl;
 		sl = (selection_list_t *) tmp->data;
+		if (!gtk_tree_row_reference_valid(sl->reference)) return;
 		addlen=strlen("smb://@://\r\n");
 		addlen += (strlen(sl->en->path));
 		addlen += ((sl->en->tag)?strlen(sl->en->tag):strlen("GUEST%%"));
@@ -772,6 +836,7 @@ void on_drag_data_get(GtkWidget * widget, GdkDragContext * context, GtkSelection
 		selection_list_t *sl;
 		char *server,*remote_file;
 		sl = (selection_list_t *) tmp->data;
+		if (!gtk_tree_row_reference_valid(sl->reference)) return;
 		server=g_strdup(sl->en->path+2);
 		strtok(server,"/");
 		if (IS_XF_NETWS(sl->en->subtype)) {
@@ -801,7 +866,8 @@ void on_drag_data_get(GtkWidget * widget, GdkDragContext * context, GtkSelection
 
     gtk_selection_data_set(selection_data, selection_data->target, 
 		    8, (const guchar *)dnd_data, selection_len); 
-
+    dragging=FALSE;
+    /*printf("DBG drag get done\n");    */
 }
 
 gboolean on_drag_motion(GtkWidget * widget, GdkDragContext * dc, gint x, gint y, guint t, gpointer data)
@@ -858,10 +924,11 @@ void on_drag_data_delete(GtkWidget * widget, GdkDragContext * context, gpointer
 
 void on_drag_end(GtkWidget * widget, GdkDragContext * context, gpointer data)
 {
-    GList *tmp = selection_list;
+    GList *tmp;
     GtkTreeView *treeview = (GtkTreeView *) widget;
     tree_details_t *tree_details = get_tree_details(treeview);
 
+    /*printf("DBG drag end\n");*/
     dragging = FALSE;
     if(!widget)
 	return;
@@ -877,24 +944,30 @@ void on_drag_end(GtkWidget * widget, GdkDragContext * context, gpointer data)
 	printf("DBG: cannot set tree_details->loading! (on_drag_end())\n");
 	return;
     }
-    while(tmp)
+    for (tmp=selection_list;tmp;tmp=tmp->next)
     {
 	selection_list_t *sl;
 	sl = (selection_list_t *) tmp->data;
-	if(sl->reference){
+	if(sl->reference && gtk_tree_row_reference_valid(sl->reference)){
 		update_dir(treeview, sl->reference);
 	}
-	g_free(sl);
-	sl=NULL;
-	tmp = tmp->next;
+    }
+
+    /* this would unselect selection after drop: */
+#if 0
+    for (tmp=selection_list;tmp;tmp=tmp->next)
+    {
+	g_free(tmp->data);
     }
     if(selection_list)
     {
 	g_list_free(selection_list);
-	selection_list = NULL;
 	selection_len = 0;
     }
     selection_list = NULL;
+#endif
+
+    
     unset_load_wait(&tree_details);
     return;
 }
@@ -905,6 +978,7 @@ void setup_drag_signal(GtkTreeView * treeview)
     g_signal_connect(G_OBJECT(treeview), "drag_data_get", G_CALLBACK(on_drag_data_get), NULL);
     g_signal_connect(G_OBJECT(treeview), "drag_motion", G_CALLBACK(on_drag_motion), NULL);
     g_signal_connect(G_OBJECT(treeview), "drag_end", G_CALLBACK(on_drag_end), NULL);
+    g_signal_connect(G_OBJECT(treeview), "drag_begin", G_CALLBACK(on_drag_begin), NULL);
 #if 0
     this signal looks bugged in gtk.only works well forsame widget.segv otherwise g_signal_connect(G_OBJECT(treeview), "drag_data_delete", G_CALLBACK(on_drag_data_delete), NULL);
 #endif
